# in progress...
import os, sys, subprocess, time
import unittest

class Main_Tests(unittest.TestCase):
    def test_basic_ops(self):  
        
        from rowordnet import RoWordNet
        from synset import Synset
        
        wn = RoWordNet()
        word = 'arbore'
        synset_ids = wn.synsets(literal=word)
        print("\nTotal number of synsets containing literal '{}': {}".format(word, len(synset_ids)))
        self.assertTrue(len(synset_ids)>0)
        
        synsets_id_nouns = wn.synsets(pos=Synset.Pos.NOUN)
        print("\tTotal number of noun synsets: {}".format(len(synsets_id_nouns)))
        self.assertTrue(len(synsets_id_nouns)>0)
    
        synsets_id_verbs = wn.synsets(pos=Synset.Pos.VERB)
        print("\tTotal number of verb synsets: {}".format(len(synsets_id_verbs)))
        self.assertTrue(len(synsets_id_verbs)>0)
        
        synsets_id_adjectives = wn.synsets(pos=Synset.Pos.ADJECTIVE)
        print("\tTotal number of adjective synsets: {}".format(len(synsets_id_adjectives)))
        self.assertTrue(len(synsets_id_adjectives)>0)
        
        synsets_id_adverbs = wn.synsets(pos=Synset.Pos.ADVERB)
        print("\tTotal number of adverb synsets: {}".format(len(synsets_id_adverbs)))
        self.assertTrue(len(synsets_id_adverbs)>0)
    
        # search for a word(here knows as a literal) in all noun synsets
        word = 'cal'
        print("\tSearch for all noun synsets that contain word/literal '{}'".format(word))    
        synset_ids = wn.synsets(literal=word, pos=Synset.Pos.NOUN)
        print("\t\tTotal number of noun synsets containing word/literal '{}' is {}".format(word, len(synset_ids)))        
        self.assertTrue(len(synset_ids)>0)
        
    def test_io(self):       
        from rowordnet import RoWordNet
        
        # load internal rowordnet
        print("\nLoading from internal resources (binary)")
        start = time.perf_counter()
        wn = RoWordNet()
        print("\t\t... done in {:.3f}s".format(time.perf_counter() - start))
        self.assertTrue(len(wn.synsets())>0)    

        # save rowordnet to xml
        print("\nSaving the rowordnet in xml file")
        start = time.perf_counter()
        wn.save("rowordnet.xml", xml=True)
        print("\t\t... done in {:.3f}s".format(time.perf_counter() - start))
        self.assertTrue(len(wn.synsets())>0)
        
        # load rowordnet from xml
        print("\nLoad the rowordnet from xml file")
        start = time.perf_counter()
        wn.load("rowordnet.xml", xml=True)
        print("\t\t... done in {:.3f}s".format(time.perf_counter() - start))
        self.assertTrue(len(wn.synsets())>0)
        
        # save rowordnet to binary
        print("\nSaving the rowordnet in binary file")
        start = time.perf_counter()
        wn.save("rowordnet.pickle")
        print("\t\t... done in {:.3f}s".format(time.perf_counter()-start))
        self.assertTrue(len(wn.synsets())>0)
        
        # load rowordnet from binary
        print("\nLoad the rowordnet from binary file")
        start = time.perf_counter()
        wn.load("rowordnet.pickle")
        print("\t\t... done in {:.3f}s".format(time.perf_counter() - start))
        self.assertTrue(len(wn.synsets())>0)
        
if __name__ == '__main__':
    """ Recreate binary from xml
    from rowordnet import RoWordNet
    wn = RoWordNet(empty=True)
    # load rowordnet from xml
    print("\n\t Load the rowordnet from xml file")
    start = time.perf_counter()
    wn.load("rowordnet.xml", xml=True)
    print("\t\t... done in {:.3f}s".format(time.perf_counter() - start))
    
    # save rowordnet to binary
    print("\n\t Saving the rowordnet in binary file")
    start = time.perf_counter()
    wn.save("rowordnet.pickle")
    print("\t\t... done in {:.3f}s".format(time.perf_counter()-start))
    """
    import os, sys
    print ("Current unit-test file location: "+os.path.realpath(__file__))
    location_of_rowordnet_package = os.path.join(os.path.dirname(os.path.dirname(os.path.realpath(__file__))),"rowordnet")
    print ("Adding "+location_of_rowordnet_package+" to path ...")
    sys.path.append(location_of_rowordnet_package)
    
    unittest.main()

    
"""
def demo_create_and_edit_synsets():
    print("\n\nThis demo shows how to create and edit synsets & relations.\n"+"_"*70)
        
    # create a synset( it's recommended to use the function 'generate_synset_id'
    # from the rowordnet class. See the function "demo_basic_rowordnet_operations'
    # for more details
    id = "my_id"
    synset = Synset(id)
    print("\n\tSynset with id '{}' has been created.".format(id))
    
    # printing the synset
    print("\n\tPrint this synset:")
    print(synset)

    # set a pos of type verb
    pos = Synset.Pos.VERB
    synset.pos = pos
    print("\tSynset's pos has been changed to '{}'". format(synset.pos))
    
    # add a literal
    literal = "tigru"
    sense = "1"
    synset.add_literal(literal=literal, sense=sense)
    print("\n\tA new literal '{}' with sense '{}' has been added to the synset with id '{}'"
          .format(literal, sense, synset.id))
    print("\tNumber of literals for synset with id '{}': {}".format(synset.id, len(synset.literals)))

    # remove a literal
    literal = "tigru"
    synset.remove_literal(literal=literal)
    print("\n\tThe literal '{}' has been removed from the synset with id '{}'".format(literal, synset.id))
    print("\tNumber of literals for synset with id '{}': {}".format(synset.id, len(synset.literals)))

    # add more literals at once
    print("\n\tAdding literals to a synset. Initially we create them:")
    literals = ['lup', 'vuple', 'caine']
    print("\tDirect addition of {} literals to synset with id '{}'".format(len(literals), synset.id))
    synset.literals = literals
    print("\tNumber of literals for synset with id '{}': {}".format(synset.id, len(synset.literals)))

    # add more senses at once 
    print("\n\tAdding senses to a synset's literals. Initially we create them:")
    literals_senses = ['1', '2', 'x']
    print("\tDirect addition of {} senses to synset with id '{}'".format(len(literals_senses), synset.id))
    synset.literals_senses = literals_senses
    print("\tNumber of senses for synset '{}': {}".format(synset.id, len(synset.literals_senses)))

    # set a definition
    definition = "Animal carnivor"
    synset.definition = definition
    print("\tSynset's defition has been changed to '{}'".format(synset.definition))

    # set a sumo
    sumo = "Animal"
    synset.sumo = sumo
    print("\tSynset's sumo has been changed to '{}'".format(synset.sumo))

    # set a sumotype
    sumotype = Synset.SumoType.INSTANCE
    synset.sumotype = sumotype
    print("\tSynset's sumotype has been changed to '{}'".format(synset.sumotype))          
          
    # generate a new id with default prefix and suffix
    wn = rowordnet.RoWordNet()
    id = wn.generate_synset_id()
    print("\n\tNew id '{}' generated with default prefix 'ENG30-' and suffix '-n'".format(id))
    # generate a new id with custom prefix and suffix
    prefix = 'ENG31-'
    suffix = '-v'
    new_id = wn.generate_synset_id(prefix=prefix, suffix=suffix)
    print("\tNew id '{}' generated with prefix '{}' and suffix '{}'".format(new_id, prefix, suffix))

    # create a synset with previous id
    synset = Synset(id)
    print("\n\tSynset with id '{}' has been created".format(synset.id))
    # add the synset to the rowordnet
    wn.add_synset(synset)
    print("\n\tAdded synset with id '{}' to the rowordnet".format(synset.id))

    # add a literal to synset
    literal = 'iepure'
    sense = '1'
    # get a synset
    synset_id = wn.synsets()[0]
    synset = wn(synset_id)
    # add a literal to the synset
    synset.add_literal(literal, sense)
    # tell the rowordnet that synsets's literals have been changed. This step is
    # necessary for a correct internal representation.
    wn.reindex_literals()
    print("\n\tAdded literal with literal '{}' and sense '{}' to the synset '{}'. "
          "Number of synsets containing literal '{}': {}"
          .format(literal, sense, synset.id, literal, len(wn.synsets(literal))))

    # remove the previous literal from synset.
    synset.remove_literal(literal)
    # again, we have to tell the rowordnet that synset's literals have been
    # changed.
    wn.reindex_literals()
    print("\tRemoved literal with literal '{}' from the synset '{}'. Number of synsets containing literal '{}': {}"
          .format(literal, synset.id, literal, len(wn.synsets(literal))))

    # generate a new synset
    prefix = 'ENG31-'
    suffix = '-n'
    new_id = wn.generate_synset_id(prefix, suffix)
    new_synset = Synset(new_id)
    wn.add_synset(new_synset)
    print("\n\tAdded new synset with id '{}' to the rowordnet".format(new_synset.id))

    # add a relation of type 'hypernym' from 'synset' to 'new_synset'
    relation = 'hypernym'
    wn.add_relation(synset.id, new_synset.id, relation)
    print("\n\tAdded '{}' relation from synset with id '{}' to synset with id '{}'"
          .format(relation, synset.id, new_synset.id))

    # remove relation of type 'hypernym' from 'synset' to 'new_synset'
    wn.remove_relation(synset.id, new_synset.id)
    print("\tRemoved relation from synset with id '{}' to synset with id '{}'".format(synset.id, new_synset.id))


def demo_operations_rowordnet():
    wn1 = rowordnet.RoWordNet()
    wn2 = rowordnet.RoWordNet()

    # add a new synset to the second wordnet
    new_synset_id = wn1.generate_synset_id()
    new_synset = Synset(new_synset_id)
    wn2.add_synset(new_synset)

    # modify a synset in the second wordnet
    synset_id = wn2.synsets("cal")[0]
    synset = wn2.synset(synset_id)
    synset.definition = "Definitie noua"

    # add a new relation in the second wordnet
    wn2.add_relation(new_synset_id, synset_id, "hypernym")

    # intersect two wordnets
    intersection_wn = intersection(wn1, wn2)
    print("Number of synsets in the intersect wordnet: {}\n".format(len(intersection_wn.synsets())))

    # merge two wordnets
    union_wn = merge(wn1, wn2)
    print("Numer of synsets in the union wordnet: {}\n".format(len(union_wn.synsets())))

    # different synsets/relations in the second wordnet
    diff_synsets, diff_relations = difference(wn1, wn2)
    print("Synsets that exists only in the second wordnet or that exists in both wordnets but are modified: {}"
          .format(diff_synsets))
    print("Relations that exists only in the second wordnet: {}".format(diff_relations))


if __name__ == '__main__':
    # rowordnet basic usage
    #demo_basic_rowordnet_operations()
    
    # rowordnet advanced usage
    #demo_get_synonymy_antonymy()
    
    # rowordnet editing
    #demo_load_and_save_rowordnet()
    #demo_create_and_edit_synsets()

    # rowordnet operations
    demo_operations_rowordnet()
    





"""


from enum import Enum
from .exceptions import SynsetError


class Synset(object):

    class Pos(Enum):
        NOUN = 0
        VERB = 1
        ADVERB = 2
        ADJECTIVE = 3

        def __str__(self):
            dic_pos2chr = {'NOUN': 'n', 'VERB': 'v', 'ADVERB': 'r', 'ADJECTIVE': 'a'}
            return dic_pos2chr[self.name]

        def __repr__(self):
            return self.name

    class SumoType(Enum):
        HYPERNYM = 0
        EQUIVALENT = 1
        INSTANCE = 2
        BRACKET = 3
        POINTS = 4

        def __str__(self):
            dic_stp2chr = {'HYPERNYM': '+', 'EQUIVALENT': '=', 'INSTANCE': '@', 'BRACKET': '[', 'POINTS': ':'}
            return dic_stp2chr[self.name]

        def __repr__(self):
            return self.name

    def __init__(self, id: str, pos=None, nonlexicalized=None, definition=None, stamp=None, sentiwn=None, domain=None,
                 sumo=None, sumotype=None, literals=None, literals_senses=None):
        """
            Initialize a synset object:

            Args:
                id (str): The id of the synset.
                pos (Pos, optional): The pos of the synset.
                nonlexicalized (str): ?
                definition (str, optional): The definition of synset.
                stamp (str, optional): The stamp of the synset.
                sentiwn (list of ints/floats, optional): The sentiwn of the synset.
                domain (str, optional): The domain of the synset.
                sumo (str, optional): The sumo of the synset.
                sumotype (SumoType, optional): The type of sumo.
                literals (dict, optional): The literals of synsets. First argument represents the literal and the second
                    one represents the sense.

            Raises:
                TypeError: If any argument has incorrect type.

        """

        if not isinstance(id, str):
            raise TypeError("Argument 'id' has incorrect type, expected str, got {}".format(type(id).__name__))

        self._id = id
        self._literals = [] if literals is None else literals
        self._literals_senses = [] if literals_senses is None else literals_senses
        self._pos = pos
        self._definition = definition
        self._stamp = stamp
        self._domain = domain
        self._sumo = sumo
        self._sumotype = sumotype
        self._sentiwn = sentiwn
        self._nonlexicalized = nonlexicalized

    @property
    def id(self):
        """
            Get/set the id(str) of this synset.
            Getter returns the id.
            Setter recieves a string containing the id.
        """
        return self._id

    @property
    def literals(self):
        """
            Get/set the literals of this synset.
            Getter returns the literals of this synset.
            Setters recieves the literals as list.
        """
        return self._literals
        
    @literals.setter
    def literals(self, value: list):
        if not isinstance(value, list):
            raise TypeError("Argument 'value' has incorrect type, expected list, got {}".format(type(value).__name__))

        for literal in value:
            if not isinstance(literal, str):
                raise TypeError("Argument 'literal-value' has incorrect type, expected str, got {}"
                                .format(type(literal).__name__))

        self._literals_senses = ["" for i in range(len(value))]
        self._literals = value

    @property
    def literals_senses(self):
        """
            Get/set the senses for each literal of this synset. Senses's indexes correspond to literals's.
            Setter returns the senses of each literal of this synset.
            Setters recieves the senses as list.
        """

        return self._literals_senses

    @literals_senses.setter
    def literals_senses(self, value: list):
        if not isinstance(value, list):
            raise TypeError("Argument 'value' has incorrect type, expected list, got {}".format(type(value).__name__))

        for sense in value:
            if not isinstance(sense, str):
                raise TypeError("Argument 'sense-value' has incorrect type, expected str, got {}"
                                .format(type(sense).__name__))

        self._literals_senses = value
           
    @property
    def sentiwn(self):
        """
            Get/set the values for the SentiWordNet(list of floats/ints) of this synset.
            Getter returns a list of 3 values for Positive, Negative, Objective.
            Setter receives a list of 3 floats/ints to set the PNO values.
        """
        return self._sentiwn
        
    @sentiwn.setter
    def sentiwn(self, value: list):
        if not isinstance(value, list):
            raise TypeError("Argument 'value' has incorrect type, expected list, got {}".format(type(value).__name__))
        if not len(value) == 3:
            raise ValueError("Argument 'value' expected a list of size 3, but got a list of size {} instead"
                             .format(len(value)))
        if not all((isinstance(element, float) or (isinstance(element, int))
                    for element in value)):
            raise ValueError("Argument's 'value' values must be of type float/int")
        if not all(0 <= element <= 1 for element in value):
            raise ValueError("Argument's 'value' values must have values between 0 and 1")
        if not sum(value) == 1:
            raise ValueError("Argument's 'value' values must add up to 1")

        self._sentiwn = value

    @property
    def definition(self):
        """
            Get/set the definition(str) of this synset.
            Getter returns the definition of this synset.
            Setter receives a string containing the definition.
        """

        return self._definition

    @definition.setter
    def definition(self, value: str):
        if not isinstance(value, str):
            raise TypeError("Argument 'value' has incorrect type, expected str, got {}".format(type(value).__name__))

        self._definition = value

    @property
    def pos(self):
        """
            Get/set the pos(Pos) of this synset.
            Getter returns the pos of this synset.
            Setter receives the pos value of this synset.
        """
        return self._pos

    @pos.setter
    def pos(self, value: Pos):
        if not isinstance(value, self.Pos):
            raise TypeError("Argument 'value' has incorrect type, expected str, got {}".format(type(value).__name__))
        self._pos = value

    @property
    def domain(self):
        """
            Gets/sets the domain of this synset.
            Getter returns the domain of this synset.
            Setter receives a string containing the domain.
        """

        return self._domain

    @domain.setter
    def domain(self, value: str):
        if not isinstance(value, str):
            raise TypeError("Argument 'value' has incorrect type, expected str, got {}".format(type(value).__name__))

        self._domain = value

    @property
    def sumo(self):
        """
            Gets/sets the sumo of this synset.
            Getter returns the sumo of this synset.
            Setter receives a string containing the sumo.
        """

        return self._sumo

    @sumo.setter
    def sumo(self, value: str):
        if not isinstance(value, str):
            raise TypeError("Argument 'value' has incorrect type, expected str, got {}".format(type(value).__name__))

        self._sumo = value

    @property
    def sumotype(self):
        """
            Gets/sets the sumotype(HYPERNYM, EQUIVALENT, INSTANCE) of this synset.
            Getter returns the sumotype of this synset.
            Setter receives a synset SumoType(HYPERNYM, EQUIVALENT, INSTANCE)
        """

        return self._sumotype

    @sumotype.setter
    def sumotype(self, value: SumoType):
        if not isinstance(value, self.SumoType):
            raise TypeError("Argument 'value' has incorrect type, expected SumoType, got {}"
                            .format(type(value).__name__))

        self._sumotype = value

    @property
    def nonlexicalized(self):
        """
            Gets/sets the nonlexicalized attribute of this synset.
            Getter returns the nonlexicalized attribute
            Setter recieves a string containing the nonlexicalized value.
        """

        return self._nonlexicalized

    @nonlexicalized.setter
    def nonlexicalized(self, value: bool):
        if not isinstance(value, bool):
            raise TypeError("Argument 'value' has incorrect type, expected bool, got {}".format(type(value).__name__))

        self._nonlexicalized = value

    @property
    def stamp(self):
        """
            Gets/sets the stamp of this synset.
            Getter returns the stamp of this synset.
            Setter recieves a string containing the stamp or None.
        """

        return self._stamp

    @stamp.setter
    def stamp(self, value: str):
        if not isinstance(value, str) and value is not None:
            raise TypeError("Argument 'value' has incorrect type, expected str, got {}".format(type(value).__name__))

        self._stamp = value

    def add_literal(self, literal: str, sense: str=""):
        """
            Add a literal to the synset.

            Args:
                literal (str): The value of the literal.
                sense (str): Sense of the literal.

            Raises:
                SynsetError: If the literal is already in the synset.
        """
        if literal in self._literals:
            raise SynsetError("Literal '{}' is already in the synset".format(literal))

        self._literals.append(literal)
        self._literals_senses.append(sense)

    def remove_literal(self, literal: str):
        """
            Remove a literal from the synset.

            Args:
                literal (str): Literal of the synset.

            Raises:
                SynsetError: If there's no literal with this value in the synset.
        """

        if literal not in self._literals:
            raise SynsetError("literal '{}' is not in the synset".format(literal))

        index = self._literals.index(literal)
        self._literals.remove(literal)
        del self._literals_senses[index]

    def __repr__(self):
        return "Synset(id={!r}, literals={!r}, definition={!r})".format(self._id, self._literals, self._definition)

    def __eq__(self, other):
        if isinstance(other, Synset):
            if (self.id == other.id and
                    self.literals == other.literals and
                    self.literals_senses == other.literals_senses and
                    self.pos == other.pos and
                    self.definition == other.definition and
                    self.stamp == other.stamp and
                    self.domain == other.domain and
                    self.sumo == other.sumo and
                    self.sumotype == other.sumotype and
                    self.sentiwn == other.sentiwn and
                    self.nonlexicalized == other.nonlexicalized):

                return True

        return False


from setuptools import setup, find_packages
# Arguments marked as "Required" below must be included for upload to PyPI.
# Fields marked as "Optional" may be commented out.

with open("README.md", "r", encoding="utf8") as fh:
    long_description = fh.read()

setup(
    # $ pip install sampleproject    
    name='rowordnet',  # Required
   
    version='1.1.0',  # Required

    description='Python API for the Romanian WordNet',  # Required

    long_description=long_description,
    long_description_content_type="text/markdown",
    maintainer = "Stefan Daniel Dumitrescu and Andrei Marius Avram",
    maintainer_email = "dumitrescu.stefan@gmail.com, avram.andreimarius@gmail.com",
    author = "Stefan Daniel Dumitrescu and Andrei Marius Avram",
    author_email = "dumitrescu.stefan@gmail.com, avram.andreimarius@gmail.com",
    
    url='https://github.com/dumitrescustefan/RoWordNet',  # Optional

    classifiers=[  # Optional        
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 5 - Production/Stable',

        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        'Intended Audience :: Information Technology',
        'Intended Audience :: Science/Research',
        
        'Topic :: Scientific/Engineering',
        'Topic :: Scientific/Engineering :: Human Machine Interfaces',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Text Processing',
        'Topic :: Text Processing :: Filters',
        'Topic :: Text Processing :: General',
        'Topic :: Text Processing :: Indexing',
        'Topic :: Text Processing :: Linguistic',
        
        'License :: OSI Approved :: GNU General Public License v3 (GPLv3)',
               
        'Programming Language :: Python :: 3'        
    ],

    keywords='romanian wordnet rowordnet rown python',  # Optional

    #packages=find_packages(exclude=['jupyter']),  # Required
    packages=find_packages("."),  # Required

    install_requires=['networkx','lxml'],  # Optional
    
    zip_safe=False,
    
    package_data={  # Optional
        'rowordnet': ['rowordnet.pickle'],
    }
)


class WordNetError(Exception):
    def __init__(self, msg):
        super(WordNetError, self).__init__(msg)


class SynsetError(Exception):
    def __init__(self, msg):
        super(SynsetError, self).__init__(msg)
























import .rowordnet
from .synset import Synset
from .rowordnet import intersection, merge, difference


def demo_basic_rowordnet_operations():
    print("\n\nThis demo shows the basic components and operations of the RoWordNet.\n"+"_"*70)
    # load rowordnet from internal resources
    wn = rowordnet.RoWordNet()
   
    # RoWordNet is composed of synsets linked together by semantic relations
    
    # the first operation is to search for a word. This will return one or more synsets.    
    word = 'arbore'
    synset_ids = wn.synsets(literal=word)
    print("\n\tTotal number of synsets containing literal '{}': {}".format(word, len(synset_ids)))
    print(synset_ids)
  
    # get a synset and print detailed information about it
    print("\n\tPrint detailed information about the first of these synsets:")
    synset_id = synset_ids[0]
    wn.print_synset(synset_id)
    
    # get the object itself
    print("\n\tGet the object itself by its id = {}, calling wn.synset():".format(synset_id))
    synset_object = wn.synset(synset_id)
    print("\t\t"+str(synset_object))
    
    print("\n\tGet the object itself by its id = {}, calling wn() directly:".format(synset_id))
    synset_object = wn(synset_id)
    print("\t\t"+str(synset_object))
    
    # print its literals, definition and id
    print("\n\tPrint its literals (synonym words): {}".format(synset_object.literals))
    print("\tPrint its definition: {}".format(synset_object.definition))
    print("\tPrint its id: {}".format(synset_object.id))
        
    # get all synsets as a list of synset IDs (list of strings)
    synsets_id = wn.synsets()
    print("\n\tTotal number of synsets: {} \n".format(len(synsets_id)))
    # example of iterating through synsets
    for synset_id in synsets_id:
        # get the synsets objects from the rowordnet by their IDs
        synset_object = wn(synset_id)
        # do something with the object
        pass

    # there are 4 types of parts of speech in RoWordNet : Nouns, Verbs, Adjectives and Adverbs
    # return all noun synsets
    synsets_id_nouns = wn.synsets(pos=Synset.Pos.NOUN)
    print("\tTotal number of noun synsets: {}".format(len(synsets_id_nouns)))
    # return all verb synsets
    synsets_id_verbs = wn.synsets(pos=Synset.Pos.VERB)
    print("\tTotal number of verb synsets: {}".format(len(synsets_id_verbs)))
    # return all adjective synsets
    synsets_id_adjectives = wn.synsets(pos=Synset.Pos.ADJECTIVE)
    print("\tTotal number of adjective synsets: {}".format(len(synsets_id_adjectives)))
    # return all adverb synsets
    synsets_id_adverbs = wn.synsets(pos=Synset.Pos.ADVERB)
    print("\tTotal number of adverb synsets: {}".format(len(synsets_id_adverbs)))
    
    # search for a word(here knows as a literal) in all noun synsets
    word = 'cal'
    print("\tSearch for all noun synsets that contain word/literal '{}'".format(word))    
    synset_ids = wn.synsets(literal=word, pos=Synset.Pos.NOUN)
    print("\t\tTotal number of noun synsets containing word/literal '{}' is {}, listed below:".format(word, len(synset_ids)))
    for synset_id in synset_ids:
        print("\t\t"+str(wn(synset_id)))
   
    # we continue with examples of navigating in the Wordnet 
    print("\n\tExamples of navigating the wordnet:")
    
    # get a synset
    synset_id = wn.synsets()[0]
    # get the path from a given synset to its root in hypermyn tree
    print("\tList of synset ids from synset with id '{}' up to its root in the hypermyn tree: ".format(synset_id))
    print("\t\t{}".format(wn.synset_to_hypernym_root(synset_id)))

    # print all outbound relations of a synset
    synset_id = wn.synsets("tren")[2]
    print("\n\tPrint all outbound relations of {}".format(wn(synset_id)))
    outbound_relations = wn.outbound_relations(synset_id)
    for outbound_relation in outbound_relations:
        target_synset_id = outbound_relation[0]        
        relation = outbound_relation[1]
        print("\t\tRelation [{}] to synset {}".format(relation, wn(target_synset_id)))
        
    # print all inbound relations of a synset, short syntax
    print("\n\tPrint all outbound relations of {}".format(wn(synset_id)))    
    for source_synset_id, relation in wn.inbound_relations(synset_id):
        print("\t\tRelation [{}] from synset {}".format(relation, wn(source_synset_id)))
        
    # get all relations of the same synset
    relations = wn.relations(synset_id)
    print("\tThe synset has {} total relations.".format(len(relations)))

    # get the shortest path between two synsets
    synset1_id = wn.synsets("cal")[2]
    synset2_id = wn.synsets("iepure")[0]    
    distance = wn.shortest_path(synset1_id, synset2_id)
    print("\n\tList of synsets containing the shortest path from synset with id '{}' to synset with id '{}': "
          .format(synset1_id, synset2_id))
    print("\t\t{}".format(distance))

    # get a new synset
    new_synset_id = wn.synsets("cal")[2]
    # travel the graph Breadth First 
    counter = 0
    print("\n\tTravel breadth-first through wordnet starting with synset '{}' (first 10 synsets) ..."
          .format(new_synset_id))
    for current_synset_id, relation, from_synset_id in wn.bfwalk(new_synset_id):
        counter += 1
        # bfwalk is a generator that yields, for each call, a BF step through wordnet 
        # you do actions with current_synset_id, relation, from_synset_id
        print("\t\t Step {}: from synset {}, with relation [{}] to synset {}".format(counter, from_synset_id, relation,
                                                                                     current_synset_id))
        if counter >= 10:
            break
        
    # get the lowest common ancestor in the hypernym tree
    synset1_id = wn.synsets("cal")[2]
    synset2_id = wn.synsets("iepure")[0]
    synset_id = wn.lowest_hypernym_common_ancestor(synset1_id, synset2_id)
    print("\n\tThe lowest common ancestor in the hypernym tree of synset: "
          "\n\t\t{} \n\t\t  and \n\t\t{} \n\t\t  is \n\t\t{}"
          .format(wn(synset1_id), wn(synset2_id), wn(synset_id)))
          
    # print all relation types existing in RoWordNet
    print("\n\tList all relation types existing in RoWordNet:")
    for relation in wn.relation_types: # this is a property
        print("\t\t{}".format(relation))


def demo_get_synonymy_antonymy():
    import itertools

    print("\n\nThis demo shows a bit more advanced series of ops.\n" + "_" * 70)

    # load from binary wordnet
    wn = rowordnet.RoWordNet()

    print("\n\tTask: We would like to extract a list of synonyms and antonyms from all the nouns in RoWordNet.")

    # get synonymy relations
    print("\n\tWe first extract synonyms directly from synsets. We list all noun synsets then iterate "
          "\n\tthrough them and create pairs from each synset.")

    synonyms = []
    synsets_id = wn.synsets()
    # for each synset, we create a list of synonyms between its literals
    for synset_id in synsets_id:
        # the literals object is a dict, but we need only the
        # actual literals (not senses)
        synset = wn(synset_id)
        literals = list(synset.literals)
        for i in range(len(literals)):
            for j in range(i+1, len(literals)):
                # append a tuple containing a pair of synonym literals
                synonyms.append((literals[i], literals[j]))

    # list a few synonyms
    print("\n\tList of the first 5 synonyms: ({} total synonym pairs extracted)".format(len(synonyms)))
    for i in range(5):
        print("\t\t {:>25} == {}".format(synonyms[i][0], synonyms[i][1]))

    # now, antonyms
    antonyms = []
    print("\n\tWe now want to extract antonyms. We look at all the antonymy relations and then for each "
          "\n\tpair of synsets in this relation we generate a cartesian product between their literals.")

    # extract all the antonymy relations from the graph and create a
    # list of synset pairs
    synset_pairs = []

    synsets_id = wn.synsets()  # extract all synsets
    for synset_id in synsets_id:
        synset = wn(synset_id)
        # extract the antonyms of a synset
        synset_outbound_id = wn.outbound_relations(synset.id)
        synset_antonyms_id = [synset_tuple[0] for synset_tuple in synset_outbound_id
                              if synset_tuple[1] == 'near_antonym']

        for synset_antonym_id in synset_antonyms_id:  # for each antonym synset
            synset_antonym = wn(synset_antonym_id)
            # if the antonymy pair doesn't already exists
            if (synset_antonym, synset) not in synset_pairs:
                # add the antonym tuple to the list
                synset_pairs.append((synset, synset_antonym))

    # for each synset pair extract its literals, so we now have a list of
    # pairs of literals
    literal_pairs = []
    for synset_pair in synset_pairs:
        # extract the literals of the first synset in the pair
        synset1_literals = list(synset_pair[0].literals)
        # extract the literals of the second synset in the pair
        synset2_literals = list(synset_pair[1].literals)
        # add a tuple containing the literals of each synset
        literal_pairs.append((synset1_literals, synset2_literals))

    # for each literals pair, we generate the cartesian product between them
    for literal_pair in literal_pairs:
        for antonym_tuple in itertools.product(literal_pair[0], literal_pair[1]):
            antonyms.append(antonym_tuple)

    # list a few antonyms
    print("\n\tList of the first 5 antonyms: ({} total antonym pairs extracted)".format(len(antonyms)))
    for i in range(5):
        print("\t\t {:>25} != {}".format(antonyms[i][0], antonyms[i][1]))


def demo_load_and_save_rowordnet():
    import time

    print("\n\nThis demo shows how to initialize, save and load a rowordnet object.\n" + "_"*70)
    
    # load internal rowordnet
    print("\n\t Loading from internal resources (binary)")
    start = time.perf_counter()
    wn = rowordnet.RoWordNet()
    print("\t\t... done in {:.3f}s".format(time.perf_counter() - start))

    # save rowordnet to xml
    print("\n\t Saving the rowordnet in xml file")
    start = time.perf_counter()
    wn.save("rowordnet.xml", xml=True)
    print("\t\t... done in {:.3f}s".format(time.perf_counter() - start))
    
    # load rowordnet from xml
    print("\n\t Load the rowordnet from xml file")
    start = time.perf_counter()
    wn.load("rowordnet.xml", xml=True)
    print("\t\t... done in {:.3f}s".format(time.perf_counter() - start))
    
    # save rowordnet to binary
    print("\n\t Saving the rowordnet in binary file")
    start = time.perf_counter()
    wn.save("rowordnet.pickle")
    print("\t\t... done in {:.3f}s".format(time.perf_counter()-start))
    
    # load rowordnet from binary
    print("\n\t Load the rowordnet from binary file")
    start = time.perf_counter()
    wn.load("rowordnet.pickle")
    print("\t\t... done in {:.3f}s".format(time.perf_counter() - start))


def demo_create_and_edit_synsets():
    print("\n\nThis demo shows how to create and edit synsets & relations.\n"+"_"*70)
        
    # create a synset( it's recommended to use the function 'generate_synset_id'
    # from the rowordnet class. See the function "demo_basic_rowordnet_operations'
    # for more details
    id = "my_id"
    synset = Synset(id)
    print("\n\tSynset with id '{}' has been created.".format(id))
    
    # printing the synset
    print("\n\tPrint this synset:")
    print(synset)

    # set a pos of type verb
    pos = Synset.Pos.VERB
    synset.pos = pos
    print("\tSynset's pos has been changed to '{}'". format(synset.pos))
    
    # add a literal
    literal = "tigru"
    sense = "1"
    synset.add_literal(literal=literal, sense=sense)
    print("\n\tA new literal '{}' with sense '{}' has been added to the synset with id '{}'"
          .format(literal, sense, synset.id))
    print("\tNumber of literals for synset with id '{}': {}".format(synset.id, len(synset.literals)))

    # remove a literal
    literal = "tigru"
    synset.remove_literal(literal=literal)
    print("\n\tThe literal '{}' has been removed from the synset with id '{}'".format(literal, synset.id))
    print("\tNumber of literals for synset with id '{}': {}".format(synset.id, len(synset.literals)))

    # add more literals at once
    print("\n\tAdding literals to a synset. Initially we create them:")
    literals = ['lup', 'vuple', 'caine']
    print("\tDirect addition of {} literals to synset with id '{}'".format(len(literals), synset.id))
    synset.literals = literals
    print("\tNumber of literals for synset with id '{}': {}".format(synset.id, len(synset.literals)))

    # add more senses at once 
    print("\n\tAdding senses to a synset's literals. Initially we create them:")
    literals_senses = ['1', '2', 'x']
    print("\tDirect addition of {} senses to synset with id '{}'".format(len(literals_senses), synset.id))
    synset.literals_senses = literals_senses
    print("\tNumber of senses for synset '{}': {}".format(synset.id, len(synset.literals_senses)))

    # set a definition
    definition = "Animal carnivor"
    synset.definition = definition
    print("\tSynset's defition has been changed to '{}'".format(synset.definition))

    # set a sumo
    sumo = "Animal"
    synset.sumo = sumo
    print("\tSynset's sumo has been changed to '{}'".format(synset.sumo))

    # set a sumotype
    sumotype = Synset.SumoType.INSTANCE
    synset.sumotype = sumotype
    print("\tSynset's sumotype has been changed to '{}'".format(synset.sumotype))          
          
    # generate a new id with default prefix and suffix
    wn = rowordnet.RoWordNet()
    id = wn.generate_synset_id()
    print("\n\tNew id '{}' generated with default prefix 'ENG30-' and suffix '-n'".format(id))
    # generate a new id with custom prefix and suffix
    prefix = 'ENG31-'
    suffix = '-v'
    new_id = wn.generate_synset_id(prefix=prefix, suffix=suffix)
    print("\tNew id '{}' generated with prefix '{}' and suffix '{}'".format(new_id, prefix, suffix))

    # create a synset with previous id
    synset = Synset(id)
    print("\n\tSynset with id '{}' has been created".format(synset.id))
    # add the synset to the rowordnet
    wn.add_synset(synset)
    print("\n\tAdded synset with id '{}' to the rowordnet".format(synset.id))

    # add a literal to synset
    literal = 'iepure'
    sense = '1'
    # get a synset
    synset_id = wn.synsets()[0]
    synset = wn(synset_id)
    # add a literal to the synset
    synset.add_literal(literal, sense)
    # tell the rowordnet that synsets's literals have been changed. This step is
    # necessary for a correct internal representation.
    wn.reindex_literals()
    print("\n\tAdded literal with literal '{}' and sense '{}' to the synset '{}'. "
          "Number of synsets containing literal '{}': {}"
          .format(literal, sense, synset.id, literal, len(wn.synsets(literal))))

    # remove the previous literal from synset.
    synset.remove_literal(literal)
    # again, we have to tell the rowordnet that synset's literals have been
    # changed.
    wn.reindex_literals()
    print("\tRemoved literal with literal '{}' from the synset '{}'. Number of synsets containing literal '{}': {}"
          .format(literal, synset.id, literal, len(wn.synsets(literal))))

    # generate a new synset
    prefix = 'ENG31-'
    suffix = '-n'
    new_id = wn.generate_synset_id(prefix, suffix)
    new_synset = Synset(new_id)
    wn.add_synset(new_synset)
    print("\n\tAdded new synset with id '{}' to the rowordnet".format(new_synset.id))

    # add a relation of type 'hypernym' from 'synset' to 'new_synset'
    relation = 'hypernym'
    wn.add_relation(synset.id, new_synset.id, relation)
    print("\n\tAdded '{}' relation from synset with id '{}' to synset with id '{}'"
          .format(relation, synset.id, new_synset.id))

    # remove relation of type 'hypernym' from 'synset' to 'new_synset'
    wn.remove_relation(synset.id, new_synset.id)
    print("\tRemoved relation from synset with id '{}' to synset with id '{}'".format(synset.id, new_synset.id))


def demo_operations_rowordnet():
    wn1 = rowordnet.RoWordNet()
    wn2 = rowordnet.RoWordNet()

    # add a new synset to the second wordnet
    new_synset_id = wn1.generate_synset_id()
    new_synset = Synset(new_synset_id)
    wn2.add_synset(new_synset)

    # modify a synset in the second wordnet
    synset_id = wn2.synsets("cal")[0]
    synset = wn2.synset(synset_id)
    synset.definition = "Definitie noua"

    # add a new relation in the second wordnet
    wn2.add_relation(new_synset_id, synset_id, "hypernym")

    # intersect two wordnets
    intersection_wn = intersection(wn1, wn2)
    print("Number of synsets in the intersect wordnet: {}\n".format(len(intersection_wn.synsets())))

    # merge two wordnets
    union_wn = merge(wn1, wn2)
    print("Numer of synsets in the union wordnet: {}\n".format(len(union_wn.synsets())))

    # different synsets/relations in the second wordnet
    diff_synsets, diff_relations = difference(wn1, wn2)
    print("Synsets that exists only in the second wordnet or that exists in both wordnets but are modified: {}"
          .format(diff_synsets))
    print("Relations that exists only in the second wordnet: {}".format(diff_relations))


if __name__ == '__main__':
    # rowordnet basic usage
    demo_basic_rowordnet_operations()
    
    # rowordnet advanced usage
    demo_get_synonymy_antonymy()
    
    # rowordnet editing
    #demo_load_and_save_rowordnet()
    #demo_create_and_edit_synsets()

    # rowordnet operations
    #demo_operations_rowordnet()

    








import pickle
import networkx as nx
import lxml.etree as et
from collections import defaultdict
from queue import Queue
import math

from .synset import Synset
from .exceptions import WordNetError


class RoWordNet(object):
    def __init__(self, filename: str = None, empty: bool = False, xml: bool = False):
        """
            Initialize a wordnet object.

            Args:
                filename (str, optional): File to load from. If filename==None, load from internal resources. Defaults
                    to None.
                empty (bool, optional): Set to True only if you want an empty wordnet object (i.e. for manual wordnet
                    editing purposes, etc.) . Defaults to False.
                xml (bool, optional): If set to True the wordnet will be loaded from an xml file. If set to False the
                    wordnet will be loaded from a binary file.
            Raises:
                TypeError: If any argument has incorrect type.
        """

        if not isinstance(filename, str) and filename is not None:
            raise TypeError("Argument 'filename' has incorrect type, expected str, got {}"
                            .format(type(filename).__name__))
        if not isinstance(empty, bool):
            raise TypeError("Argument 'empty' has incorrect type, expected bool, got {}".format(type(empty).__name__))
        if not isinstance(xml, bool):
            raise TypeError("Argument 'xml' has incorrect type, expected bool, got {}".format(type(xml).__name__))

        self._clean()
        if empty:
            return

        if filename is None:
            import pkg_resources
            path = "rowordnet.pickle"  # always use slash
            filepath = pkg_resources.resource_filename(__name__, path)
            self._load_from_binary(filepath)
        else:
            if xml is True:
                self._load_from_xml(filename)
            else:
                self._load_from_binary(filename)

        self._max_hypernym_height = self._hypernym_tree_height("ENG30-00002684-n")

    def _clean(self):
        self._graph = nx.DiGraph()
        self._hypernym_graph = nx.DiGraph()
        self._synsets = {}
        self._literal2synset = defaultdict(list)
        self._literal2synset_strict = defaultdict(list)
        self._relation_types = set()

    @property
    def relation_types(self):
        """
            Gets a set of all relation types existing in RoWordNet.
            Returns:
                set of str: A set containing all types of existing relations between synsets.
        """

        return self._relation_types

    def save(self, filename: str, xml: bool = False):
        """
            Save a wordnet object in a given file.

            Args:
                filename (str): The file where the wordnet will be saved.
                xml (bool, optional): If set to True, it will save in xml format. If set to False it will save in binary
                    format. Defaults to False.
            Raises:
                TypeError: If any argument has incorrect type.
        """

        if not isinstance(filename, str):
            raise TypeError("Argument 'filename' has incorrect type, expected str, got {}"
                            .format(type(filename).__name__))
        if not isinstance(xml, bool):
            raise TypeError("Argument 'xml' has incorrect type, expected bool, got {}".format(type(xml).__name__))

        if xml is True:
            self._save_to_xml(filename)
        elif xml is False:
            self._save_to_binary(filename)

    def load(self, filename: str, xml: bool = False):
        """
            Load a wordnet object from a given file.
            Args:
                filename (str): The file from where wordnet will be loaded.
                xml (bool, optional): If set to True, it will load from xml format. If set to False, it will load from
                    binary format. Defaults to False.
            Raises:
                TypeError: If any argument has incorrect type.
        """

        if not isinstance(filename, str):
            raise TypeError("Argument 'filename' has incorrect type, expected str, got {}"
                            .format(type(filename).__name__))
        if not isinstance(xml, bool):
            raise TypeError("Argument 'xml' has incorrect type, expected bool, got {}".format(type(xml).__name__))

        if xml is True:
            self._load_from_xml(filename)
        elif xml is False:
            self._load_from_binary(filename)

    def _load_from_xml(self, filename: str):
        self._clean()

        parser = et.XMLParser(encoding="utf-8")
        root = et.parse(filename, parser).getroot()

        for child in root:
            synset = None

            for element in child:
                if element.tag == 'ID':
                    synset = Synset(element.text)

                if element.tag == 'POS':
                    dic_chr2pos = {
                        'n': Synset.Pos.NOUN,
                        'v': Synset.Pos.VERB,
                        'r': Synset.Pos.ADVERB,
                        'a': Synset.Pos.ADJECTIVE
                    }
                    pos = dic_chr2pos[element.text]
                    synset.pos = pos

                if element.tag == 'SYNONYM':
                    try:
                        synset.literals = [literal.text for literal in element]
                    except TypeError as e:
                        print(synset.id)

                    literals_senses = []
                    for literal in element:
                        literals_senses.append(literal[0].text if literal[0].text is not None else "")
                    synset.literals_senses = literals_senses

                    # index literals
                    for literal in synset.literals:
                        self._literal2synset[literal].append(synset.id)
                        self._literal2synset_strict[literal].append(synset.id)
                        literal_parts = literal.split('_')
                        if len(literal_parts) > 1:  # add composing words for multi-word literals in non-strict index
                            for literal_part in literal_parts:
                                self._literal2synset[literal_part].append(synset.id)

                if element.tag == 'STAMP':
                    synset.stamp = element.text

                if element.tag == 'ILR':
                    self._relation_types.add(element[0].text)

                    self._graph.add_edge(synset.id, element.text,
                                         label=element[0].text)

                if element.tag == 'DEF':
                    synset.definition = element.text

                if element.tag == 'DOMAIN':
                    synset.domain = element.text

                if element.tag == 'SUMO':
                    synset.sumo = element.text
                    dic_chr2sumotype = {
                        '+': Synset.SumoType.HYPERNYM,
                        '=': Synset.SumoType.EQUIVALENT,
                        '@': Synset.SumoType.INSTANCE,
                        '[': Synset.SumoType.BRACKET,
                        ':': Synset.SumoType.POINTS
                    }
                    sumotype = dic_chr2sumotype[element[0].text]
                    synset.sumotype = sumotype

                if element.tag == 'SENTIWN':
                    synset.sentiwn = [float(subelement.text) for subelement in element]

            self._synsets[synset.id] = synset

    def _load_from_binary(self, filename: str):
        with open(filename, "rb") as f:
            wn = pickle.load(f)

        self._clean()

        self._relation_types = wn.relation_types

        synsets_id = wn.synsets()

        for synset_id in synsets_id:
            self._graph.add_node(synset_id)
            self._hypernym_graph.add_node(synset_id)

            adj_synsets_id = wn.outbound_relations(synset_id)
            for adj_synset_id, relation in adj_synsets_id:
                self._graph.add_edge(synset_id, adj_synset_id, label=relation)

                if relation == "hypernym" or relation == "hyponym":
                    self._hypernym_graph.add_edge(synset_id, adj_synset_id, label=relation)

        for synset_id in synsets_id:
            self._synsets[synset_id] = wn.synset(synset_id)

        for synset_id in synsets_id:
            synset = wn.synset(synset_id)
            for literal in synset.literals:
                self._literal2synset[literal].append(synset.id)
                self._literal2synset_strict[literal].append(synset.id)
                literal_parts = literal.split('_')
                if len(literal_parts) > 1:  # add composing words for multi-word literals in non-strict index
                    for literal_part in literal_parts:
                        self._literal2synset[literal_part].append(synset.id)

    def _save_to_xml(self, filename: str):
        root = et.Element("ROWN")

        for synset in self._synsets.values():
            syn = et.SubElement(root, "SYNSET")

            et.SubElement(syn, "ID").text = synset.id

            et.SubElement(syn, "POS").text = str(synset.pos)

            synonym = et.SubElement(syn, "SYNONYM")
            for literal in synset.literals:
                lit = et.SubElement(synonym, "LITERAL")
                lit.text = literal
                literal_index = synset.literals.index(literal)
                literal_sense = synset.literals_senses[literal_index]
                et.SubElement(lit, "SENSE").text = literal_sense

            if synset.stamp is not None:
                et.SubElement(syn, "STAMP").text = synset.stamp

            if synset.id in self._graph.adj.keys():
                for target_node_id, edge_data in self._graph[synset.id].items():
                    ilr = et.SubElement(syn, "ILR")
                    ilr.text = target_node_id
                    et.SubElement(ilr, "TYPE").text = edge_data['label']
            if synset.definition is not None:
                et.SubElement(syn, "DEF").text = synset.definition

            if synset.domain is not None:
                et.SubElement(syn, "DOMAIN").text = synset.domain

            if synset.sumo is not None:
                sumo = et.SubElement(syn, "SUMO")
                sumo.text = synset.sumo
                et.SubElement(sumo, "TYPE").text = str(synset.sumotype)

            if synset.sentiwn is not None:
                sentiwn = et.SubElement(syn, "SENTIWN")
                et.SubElement(sentiwn, "P").text = str(synset.sentiwn[0])
                et.SubElement(sentiwn, "N").text = str(synset.sentiwn[1])
                et.SubElement(sentiwn, "O").text = str(synset.sentiwn[2])

        tree = et.ElementTree(root)
        tree.write(filename, encoding="utf-8", pretty_print=True)

    def _save_to_binary(self, filename: str):
        with open(filename, "wb") as f:
            pickle.dump(self, f)

    def synsets(self, literal: str = None, pos: Synset.Pos = None, strict: bool = False):
        """
            Get a list of synsets. If a literal is given, only the synsets that contain that literal will be selected.
            If a pos is given, only the synsets that have that pos will be selected.
            Example: searching for "tren", if strict is False, return "tren_de_aterizare"; if strict is True, skip the
            multi-word expression literal and return exact matches (i.e. only "tren").
            Args:
                literal (str, optional): The literal that synsets must contain. Defaults to None.
                pos (Synset.Pos, optional): The type of pos that synsets must have. Defaults to None.
                strict (bool, optional): Retrieve exact results. Defaults to False.
            Returns:
                list of Synsets: A list containing the desired synsets. If no synset with the given word is found, it
                will return an empty list.
            Raises:
                TypeError: If any argument has incorrect type.
        """

        if literal is None:
            synsets_id = list(self._synsets.keys())
        else:
            if not isinstance(literal, str):
                raise TypeError("Argument 'literal' has incorrect type, expected str, got {}"
                                .format(type(literal).__name__))

            if strict:
                if literal not in self._literal2synset_strict:
                    return []
                synsets_id = self._literal2synset_strict[literal]
            else:
                if literal not in self._literal2synset:
                    return []
                synsets_id = self._literal2synset[literal]

        if pos is not None:
            if not isinstance(pos, Synset.Pos):
                raise TypeError("Argument 'pos' has incorrect type, expected Synset.Pos, got {}"
                                .format(type(pos).__name__))

            synsets_id = [synset_id for synset_id in synsets_id if self._synsets[synset_id].pos == pos]

        return synsets_id

    def print_synset(self, synset_id: str):
        """
            Fully prints a synset.

            Args:
                synset_id(str): Id of the synset.

            Raises:
                TypeError: If any argument has incorrect type.
                WordNetError: If there's no synset with the given id in the wordnet.
        """

        if not isinstance(synset_id, str):
            raise TypeError("Argument 'synset_id' has incorrect type, expected str, got {}"
                            .format(type(synset_id).__name__))
        if synset_id not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id))

        synset = self.synset(synset_id)

        print("Synset: "
              "\n\t  id={}"
              "\n\t  pos={!r}"
              "\n\t  nonlexicalized={}"
              "\n\t  stamp={}"
              "\n\t  domain={}"
              "\n\t  definition={}" 
              "\n\t  sumo={} "
              "\n\t  sumoType={!r}"
              "\n\t  sentiwn={}".
              format(synset.id, synset.pos, synset.nonlexicalized, synset.stamp, synset.domain, synset.definition,
                     synset.sumo, synset.sumotype, synset.sentiwn))

        print("\t  Literals:")
        for i in range(len(synset.literals)):
            print("\t\t  {} - {}".format(synset.literals[i], synset.literals_senses[i]))

        outbound_relations = self.outbound_relations(synset_id)
        print("\t  Outbound relations: ")
        for out_synset_id, relation in outbound_relations:
            print("\t\t  {} - {}".format(out_synset_id, relation))

        inbound_relations = self.inbound_relations(synset_id)
        print("\t  Inbound relations: ")
        for in_synset_id, relation in inbound_relations:
            print("\t\t  {} - {}".format(in_synset_id, relation))

    def synset_exists(self, synset: Synset):
        if not isinstance(synset, Synset):
            raise TypeError("Argument 'synset' has incorrect type, expected Synset, got {}"
                            .format(type(synset).__name__))
        if synset.id not in self._synsets.keys():
            return False

        current_synset = self._synsets[synset.id]
        if current_synset != synset:
            return False

        return True

    def reindex_literals(self):
        """
            Reindex all literals to the synsets. This is used if the literals of a synset have been changed.
        """

        self._literal2synset.clear()
        self._literal2synset_strict.clear()
        for synset in self._synsets.values():
            for literal in synset.literals:
                self._literal2synset[literal].append(synset.id)
                [literal].append(synset.id)
                literal_parts = literal.split('_')
                if len(literal_parts) > 1:  # add composing words for multi-word literals in non-strict index
                    for literal_part in literal_parts:
                        self._literal2synset[literal_part].append(synset.id)

    def inbound_relations(self, synset_id: str):
        if not isinstance(synset_id, str):
            raise TypeError("Argument 'synset_id' has incorrect type, expected str, got {}"
                            .format(type(synset_id).__name__))
        if synset_id not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id))

        inbound_relations = []
        for synset_id_iter in self._graph.adj.keys():
            for adj_synset_id, data in self._graph.adj[synset_id_iter].items():
                if adj_synset_id == synset_id:
                    inbound_relations.append((synset_id_iter, data['label']))

        return inbound_relations

    def outbound_relations(self, synset_id: str):
        if not isinstance(synset_id, str):
            raise TypeError("Argument 'synset_id' has incorrect type, expected str, got {}"
                            .format(type(synset_id).__name__))
        if synset_id not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id))

        if synset_id not in self._graph.adj.keys():
            return []

        outbound_relations = []
        for adj_synset_id, data in self._graph.adj[synset_id].items():
            outbound_relations.append((adj_synset_id, data['label']))

        return outbound_relations

    def relations(self, synset_id: str):
        return self.outbound_relations(synset_id) + self.inbound_relations(synset_id)

    def relation_exists(self, synset_id1: str, synset_id2: str, relation: str):
        if not isinstance(synset_id1, str):
            raise TypeError("Argument 'synset_id1' has incorrect type, expected str, got {}"
                            .format(type(synset_id1).__name__))
        if not isinstance(synset_id2, str):
            raise TypeError("Argument 'synset_id2' has incorrect type, expected str, got {}"
                            .format(type(synset_id2).__name__))
        if not isinstance(relation, str):
            raise TypeError("Argument 'relation' has incorrect type, expected str, got {}"
                            .format(type(relation).__name__))
        if synset_id1 not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id1))
        if synset_id2 not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id2))
        if relation not in self._relation_types:
            raise WordNetError("Relation '{}' is not a correct relation".format(relation))

        for adj_synset_id, data in self._graph.adj[synset_id1].items():
            if adj_synset_id == synset_id2 and data['label'] == relation:
                return True

        return False

    def add_relation_type(self, relation_type: str):
        if not isinstance(relation_type, str):
            raise TypeError("Argument 'relation_type' has incorrect type, expected str, got {}"
                            .format(type(relation_type).__name__))
        if relation_type in self._relation_types:
            raise WordNetError("Relation type {} is already in the wordnet".format(relation_type))

        self._relation_types.add(relation_type)

    def __call__(self, synset_id: str):
        if not isinstance(synset_id, str):
            raise TypeError("Argument 'synset_id' has incorrect type, expected str, got {}"
                            .format(type(synset_id).__name__))
        if synset_id not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id))

        return self.synset(synset_id)
        
    def synset(self, synset_id: str):
        """
            Get a synset, given its id.
            Args:
                synset_id (str): The id of the synset.
            Returns:
                Synset: The synset with the desired id. If no synset is found, the function will return None.
            Raises:
                TypeError: If any argument has incorrect type.
                WordNetError: If there's no synset with the given id in the wordnet.
        """

        if not isinstance(synset_id, str):
            raise TypeError("Argument 'synset_id' has incorrect type, expected str, got {}"
                            .format(type(synset_id).__name__))
        if synset_id not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id))

        return self._synsets[synset_id]

    def generate_synset_id(self, prefix: str = 'ENG30-', suffix: str = '-n'):
        """
            Generate the first available id that starts with the given prefix and ends with the given suffix.
            Args:
                prefix (str, optional): The desired prefix. Defaults to 'ENG30-'
                suffix (str, optional): The desired suffix. Defaults to '-n'.
            Returns:
                str: The first available id that starts with the given prefix and ends with the given suffix.
            Raises:
                TypeError: If any argument has incorrect type.
        """

        if not isinstance(suffix, str):
            raise TypeError("Argument 'suffix' has incorrect type, expected str, got {}"
                            .format(type(suffix).__name__))
        if not isinstance(prefix, str):
            raise TypeError("Argument 'prefix' has incorrect type, expected str, got {}"
                            .format(type(prefix).__name__))

        matched = False
        maximum = 0

        for synset_id in self._synsets:
            if not synset_id.startswith(prefix):
                continue
            if not synset_id.endswith(suffix):
                continue

            matched = True

            synset_id = synset_id.replace(prefix, "")
            synset_id = synset_id.replace(suffix, "")

            if int(synset_id) > maximum:
                maximum = int(synset_id)

        if matched:
            return "{}{:08d}{}".format(prefix, maximum + 1, suffix)
        else:
            return "{}{:08d}{}".format(prefix, 1, suffix)

    def add_synset(self, synset: Synset):
        """
            Add a synset to wordnet.
            Args:
                 synset (Synset): The synset to be added.
            Raises:
                TypeError: If any argument has incorrect type.
                WordNetError: If a synset with the given id is already in the wordnet.
        """

        if not isinstance(synset, Synset):
            raise TypeError("Argument 'synset' has incorrect type, expected Synset, got {}"
                            .format(type(synset).__name__))
        if synset.id in self._synsets:
            raise WordNetError("Synset with id '{}' is already in the wordnet".format(synset.id))

        self._graph.add_node(synset.id)
        self._synsets[synset.id] = synset
        for literal in synset.literals:
            self._literal2synset[literal].append(synset.id)
            self._literal2synset_strict[literal].append(synset.id)
            literal_parts = literal.split('_')
            if len(literal_parts) > 1:  # add composing words for multi-word literals in non-strict index
                for literal_part in literal_parts:
                    self._literal2synset[literal_part].append(synset.id)

    def add_relation(self, synset_id1: str, synset_id2: str, relation: str):
        """
            Add a new relation to the wordnet. Relation will always be from
            synset_id1 to synset_id2.
            Args:
                synset_id1 (str): Id of the first synset.
                synset_id2 (str): Id of the second synset.
                relation(str): Relation type between the synsets.
            Raises:
                TypeError: If any argument has incorrect type.
                WordNetError: If there's no synset with the given ids in the wordnet, if there's already a relation from
                    the first synset to the second synset or if the given relation has an incorrect value.
        """

        if not isinstance(synset_id1, str):
            raise TypeError("Argument 'synset_id1' has incorrect type, expected str, got {}"
                            .format(type(synset_id1).__name__))
        if not isinstance(synset_id2, str):
            raise TypeError("Argument 'synset_id2' has incorrect type, expected str, got {}"
                            .format(type(synset_id2).__name__))
        if not isinstance(relation, str):
            raise TypeError("Argument 'relation' has incorrect type, expected str, got {}"
                            .format(type(relation).__name__))
        if synset_id1 not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id1))
        if synset_id2 not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id2))
        if relation not in self._relation_types:
            raise WordNetError("Relation '{}' is not a correct relation".format(relation))
        if self._graph.has_edge(synset_id1, synset_id2):
            raise WordNetError("There's already a relation from the synset with id '{}' to the synset with id '{}'"
                               .format(synset_id1, synset_id2))

        self._graph.add_edge(synset_id1, synset_id2, label=relation)

    def remove_relation(self, synset_id1: str, synset_id2: str):
        """
            Remove a relation between two synsets. Relation is always from the first synset to the second synset.
            Args:
                synset_id1 (str): Id of the first synset.
                synset_id2 (str): Id of the second synset.
            Raises:
                TypeError: If any argument has incorrect type.
                WordNetError: If there's no synset with the given ids in the wordnet or if there's no relation from the
                    first synset to the second synset.
        """

        if not isinstance(synset_id1, str):
            raise TypeError("Argument 'synset_id1' has incorrect type, expected str, got {}"
                            .format(type(synset_id1).__name__))
        if not isinstance(synset_id2, str):
            raise TypeError("Argument 'synset_id2' has incorrect type, expected str, got {}"
                            .format(type(synset_id2).__name__))
        if synset_id1 not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id1))
        if synset_id2 not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id2))
        if not self._graph.has_edge(synset_id1, synset_id2):
            raise WordNetError("There's no relation from the synset with id '{}' to the synset with id '{}'"
                               .format(synset_id1, synset_id2))

        self._graph.remove_edge(synset_id1, synset_id2)

    def synset_to_hypernym_root(self, synset_id: str):
        """
            Get a list containing the path from the given synset to the root in a specified tree.
            Args:
                synset_id (str): Id of the synset.
            Returns:
                list: A list containing synset ids that create the path to the root of the tree.
            Raises:
                TypeError: If any argument has incorrect type.
                WordNetError: If there's no synset with the given id in the wordnet.
        """

        if not isinstance(synset_id, str):
            raise TypeError("Argument 'synset_id' has incorrect type, expected str, got {}"
                            .format(type(synset_id).__name__))
        if synset_id not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id))

        synset_id_ancestor = synset_id
        synset_id_to_root = [synset_id]

        while synset_id_ancestor is not None:
            adj_synsets = self._hypernym_graph[synset_id_ancestor]
            synset_id_ancestor = None

            for adj_synset_id, adj_synset_data in adj_synsets.items():
                if adj_synset_data['label'] == 'hypernym':
                    synset_id_to_root.append(adj_synset_id)
                    synset_id_ancestor = adj_synset_id
                    break

        return synset_id_to_root

    def lowest_hypernym_common_ancestor(self, synset_id1: str, synset_id2: str):
        """
            Find the lowest common ancestor of two synsets in a specified tree.
            Args:
                synset_id1 (str): Id of the first synset.
                synset_id2 (str): Id of the second synset.
            Returns:
                str: A synset ID representing the lowest common ancestor in the specified tree.
            Raises:
                TypeError: If any argument has incorrect type.
                WordNetError: If there's no synset with the given ids in the wordnet.value.
        """

        synset_id1_to_root = self.synset_to_hypernym_root(synset_id1)
        synset_id2_to_root = self.synset_to_hypernym_root(synset_id2)

        for synset_1 in synset_id1_to_root:
            for synset_2 in synset_id2_to_root:
                if synset_1 == synset_2:
                    lowest_common_ancestor = synset_1
                    return lowest_common_ancestor

    def bfwalk(self, synset_id: str):
        """
            Travel the wordnet starting from a given synset.
            Args:
                synset_id (str): The id of the synset.
            Yields:
                Synset: The next synset in the wordnet.
            Raises:
                TypeError: If any argument has incorrect type.
                WordNetError: If there's no synset with the given id in the wordnet.
        """

        if not isinstance(synset_id, str):
            raise TypeError("Argument 'synset_id' has incorrect type, expected str, got {}"
                            .format(type(synset_id).__name__))
        if synset_id not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id))

        queue = Queue()
        marked_synsets_id = [synset_id]
        from_synsets_rel = dict()

        for adj_synset_id, data in self._graph.adj[synset_id].items():
            from_synsets_rel[adj_synset_id] = (data['label'], synset_id)
            queue.put(adj_synset_id)
            marked_synsets_id.append(adj_synset_id)

        while not queue.empty():
            cur_synset_id = queue.get()

            adj_synsets_id = self._graph.adj[cur_synset_id]

            for adj_synset_id, data in adj_synsets_id.items():
                if adj_synset_id not in marked_synsets_id:
                    marked_synsets_id.append(adj_synset_id)
                    queue.put(adj_synset_id)
                    from_synsets_rel[adj_synset_id] = (data['label'], cur_synset_id)

            yield cur_synset_id, from_synsets_rel[cur_synset_id][0], from_synsets_rel[cur_synset_id][1]

            from_synsets_rel.pop(cur_synset_id)

    def shortest_path(self, synset_id1: str, synset_id2: str, relations: set = None):
        """
            Get the shortest path from the first synset to the second synset.
            Args:
                synset_id1 (str): Id of the first synset.
                synset_id2 (str): Id of the second synset.
                relations (list of str): The allowed relations in the shortest path algorithm.
            Returns:
                list of str: A list of synset ids representing the path from
                the first synset to the second synset.
            Raises:
                TypeError: If any argument has incorrect type.
                WordNetError: If there's no synset with the given ids in the wordnet or if any relation has an incorrect
                    value.
                networkx.exception.NetworkXNoPath: If there is no path between the synsets.
        """

        if not isinstance(synset_id1, str):
            raise TypeError("Argument 'synset_id1' has incorrect type, expected str, got {}"
                            .format(type(synset_id1).__name__))
        if not isinstance(synset_id2, str):
            raise TypeError("Argument 'synset_id2' has incorrect type, expected str, got {}"
                            .format(type(synset_id2).__name__))
        if synset_id1 not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id1))
        if synset_id2 not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id2))

        if relations is None:
            return nx.shortest_path(self._graph, synset_id1, synset_id2)
        else:
            if not isinstance(relations, set):
                raise TypeError("Argument 'relations' has incorrect type, expected set, got {}"
                                .format(type(relations).__name__))

            for relation in relations:
                if not isinstance(relation, str):
                    raise TypeError("Argument 'relation - relations' has incorrect type, expected str, got {}"
                                    .format(type(relation).__name__))
                if relation not in self._relation_types:
                    raise WordNetError("Relation '{}' is not a correct relation".format(relation))

            if relations == {"hypernym", "hyponym"}:
                return nx.shortest_path(self._hypernym_graph, synset_id1, synset_id2)
            else:
                raise NotImplemented("The current set of relations is not supported anymore by the function.")

    def path_similarity(self, synset_id1: str, synset_id2: str, simulate_root: bool = True):
        """
            Returns the path similarity between two synsets.
            Args:
                synset_id1 (str): Id of the first synset.
                synset_id2 (str): Id of the second synset.
                simulate_root (bool): Simulate a virtual root if there is no common root for the two synsets.
            Returns:
                float: 1 if the synsets are equal, None if there is no path between the synsets or
                       1 / 1/(shortest_path_distance + 1) otherwise.
            Raises:
                TypeError: If any argument has incorrect type.
                WordNetError: If there's no synset with the given ids in the wordnet or if any relation has an incorrect
                    value.
        """
        if not isinstance(synset_id1, str):
            raise TypeError("Argument 'synset_id1' has incorrect type, expected str, got {}"
                            .format(type(synset_id1).__name__))
        if not isinstance(synset_id2, str):
            raise TypeError("Argument 'synset_id2' has incorrect type, expected str, got {}"
                            .format(type(synset_id2).__name__))

        if synset_id1 not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id1))
        if synset_id2 not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id2))

        if synset_id1 == synset_id2:
            return 1

        try:
            shortest_path_distance = len(self.shortest_path(synset_id1, synset_id2, relations={"hypernym", "hyponym"}))
        except nx.exception.NetworkXNoPath:
            if simulate_root:
                depth_synset1 = len(self.synset_to_hypernym_root(synset_id1))
                depth_synset2 = len(self.synset_to_hypernym_root(synset_id2))

                shortest_path_distance = depth_synset1 + depth_synset2 + 2
            else:
                return None

        return 1 / (shortest_path_distance + 1)

    def wup_similarity(self, synset_id1: str, synset_id2: str, simulate_root: bool = True):
        """
           Returns the Wu and Palmer similarity between two synsets.
           Args:
               synset_id1 (str): Id of the first synset.
               synset_id2 (str): Id of the second synset.
               simulate_root (bool): Simulate a virtual root if there is no common root for the two synsets.
           Returns:
               float: None if there is no path between the synsets or distance between synsets defined as
                      2 * depth(least_common_subsumer(synset1,synset2))) / (depth(synset1) + depth(synset2)) otherwise.
           Raises:
               TypeError: If any argument has incorrect type.
               WordNetError: If there's no synset with the given ids in the wordnet or if any relation has an incorrect
                   value.
        """
        if not isinstance(synset_id1, str):
            raise TypeError("Argument 'synset_id1' has incorrect type, expected str, got {}"
                            .format(type(synset_id1).__name__))
        if not isinstance(synset_id2, str):
            raise TypeError("Argument 'synset_id2' has incorrect type, expected str, got {}"
                            .format(type(synset_id2).__name__))

        if synset_id1 not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id1))
        if synset_id2 not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id2))

        if synset_id1[-1] != synset_id2[-1]:
            return None

        lcs_synset = self.lowest_hypernym_common_ancestor(synset_id1, synset_id2)

        depth_synset1 = len(self.synset_to_hypernym_root(synset_id1))
        depth_synset2 = len(self.synset_to_hypernym_root(synset_id2))

        if lcs_synset is None:
            if simulate_root:
                depth_synset1 += 1
                depth_synset2 += 1
                depth_lcs_synset = 1
            else:
                return None
        else:
            depth_lcs_synset = len(self.synset_to_hypernym_root(lcs_synset))

        return 2 * depth_lcs_synset / (depth_synset1 + depth_synset2)

    def _hypernym_tree_height(self, root_id):
        depths = []

        for adj_synset_id, data in self._graph.adj[root_id].items():
            if data['label'] == 'hyponym':
                depths.append(self._hypernym_tree_height(adj_synset_id))

        if len(depths) == 0:
            return 0

        return max(depths) + 1

    def lch_similarity(self, synset_id1: str, synset_id2: str, simulate_root: bool = True):
        """
            Returns the Leacock and Chodorow similarity between two synsets.
            Args:
                synset_id1 (str): Id of the first synset.
                synset_id2 (str): Id of the second synset.
                simulate_root (bool): Simulate a virtual root if there is no common root for the two synsets.
            Returns:
                float: None if there is no path between the synsets or distance between synsets defined as
                       -log ((dist(synset1,synset2)+1) / (2 * maximum taxonomy depth)) otherwise.
            Raises:
                TypeError: If any argument has incorrect type.
                WordNetError: If there's no synset with the given ids in the wordnet or if any relation has an incorrect
                    value.
        """
        if not isinstance(synset_id1, str):
            raise TypeError("Argument 'synset_id1' has incorrect type, expected str, got {}"
                            .format(type(synset_id1).__name__))
        if not isinstance(synset_id2, str):
            raise TypeError("Argument 'synset_id2' has incorrect type, expected str, got {}"
                            .format(type(synset_id2).__name__))

        if synset_id1 not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id1))
        if synset_id2 not in self._synsets:
            raise WordNetError("Synset with id '{}' is not in the wordnet".format(synset_id2))

        max_hypernym_height = self._max_hypernym_height

        try:
            shortest_path_distance = len(self.shortest_path(synset_id1, synset_id2, relations={"hypernym", "hyponym"}))
        except nx.exception.NetworkXNoPath:
            if simulate_root:
                depth_synset1 = len(self.synset_to_hypernym_root(synset_id1))
                depth_synset2 = len(self.synset_to_hypernym_root(synset_id2))

                shortest_path_distance = depth_synset1 + depth_synset2 + 2

                max_hypernym_height += 1
            else:
                return None

        return - math.log2((shortest_path_distance + 1) / (2 * max_hypernym_height))


def intersection(wordnet_1, wordnet_2):
    if not isinstance(wordnet_1, RoWordNet):
        raise TypeError("Argument 'wordnet_1' has incorrect type, expected RoWordNet, got {}"
                        .format(type(wordnet_1).__name__))
    if not isinstance(wordnet_2, RoWordNet):
        raise TypeError("Argument 'wordnet_2' has incorrect type, expected RoWordNet, got {}"
                        .format(type(wordnet_2).__name__))

    intersection_wordnet = RoWordNet(empty=True)

    for synset_id in wordnet_1.synsets():
        synset = wordnet_1.synset(synset_id)
        if wordnet_2.synset_exists(synset):
            intersection_wordnet.add_synset(synset)

    relations_type_wn1 = wordnet_1.relation_types
    relations_type_wn2 = wordnet_2.relation_types

    for relation_type_wn1 in relations_type_wn1:
        for relation_type_wn2 in relations_type_wn2:
            if relation_type_wn1 == relation_type_wn2:
                intersection_wordnet.add_relation_type(relation_type_wn1)
                break

    for synset_id in wordnet_1.synsets():
        relations_wn1 = wordnet_1.outbound_relations(synset_id)
        for relation_wn1 in relations_wn1:
            try:
                intersection_wordnet.add_relation(synset_id, relation_wn1[0], relation_wn1[1])
            except WordNetError:
                pass

    return intersection_wordnet


def merge(wordnet_1: RoWordNet, wordnet_2: RoWordNet):
    if not isinstance(wordnet_1, RoWordNet):
        raise TypeError("Argument 'wordnet_1' has incorrect type, expected RoWordNet, got {}"
                        .format(type(wordnet_1).__name__))
    if not isinstance(wordnet_2, RoWordNet):
        raise TypeError("Argument 'wordnet_2' has incorrect type, expected RoWordNet, got {}"
                        .format(type(wordnet_2).__name__))

    new_wordnet = RoWordNet(empty=True)

    # copying the second wordnet
    for synset_id in wordnet_2.synsets():
        synset = wordnet_2.synset(synset_id)
        new_wordnet.add_synset(synset)

    relations_type = wordnet_2.relation_types
    for relation_type in relations_type:
        new_wordnet.add_relation_type(relation_type)

    for synset_id in wordnet_2.synsets():
        relations = wordnet_2.outbound_relations(synset_id)
        for relation in relations:
            new_wordnet.add_relation(synset_id, relation[0], relation[1])

    # add the first wordnet
    relations_type = wordnet_1.relation_types
    for relation_type in relations_type:
        try:
            new_wordnet.add_relation_type(relation_type)
        except WordNetError:
            pass

    for synset_id in wordnet_1.synsets():
        try:
            synset = wordnet_1.synset(synset_id)
            new_wordnet.add_synset(synset)
        except WordNetError:
            pass

    for synset_id in wordnet_1.synsets():
        relations = wordnet_1.outbound_relations(synset_id)
        for relation in relations:
            try:
                new_wordnet.add_relation(synset_id, relation[0], relation[1])
            except WordNetError:
                pass

    return new_wordnet


def difference(wordnet_1, wordnet_2):
    if not isinstance(wordnet_1, RoWordNet):
        raise TypeError("Argument 'wordnet_1' has incorrect type, expected RoWordNet, got {}"
                        .format(type(wordnet_1).__name__))
    if not isinstance(wordnet_2, RoWordNet):
        raise TypeError("Argument 'wordnet_2' has incorrect type, expected RoWordNet, got {}"
                        .format(type(wordnet_2).__name__))

    diff_synsets = set()
    diff_relations = set()

    for synset_id in wordnet_2.synsets():
        try:
            synset1 = wordnet_1.synset(synset_id)
            synset2 = wordnet_2.synset(synset_id)
            if synset1 != synset2:
                diff_synsets.add(synset_id)
        except WordNetError:
            diff_synsets.add(synset_id)

    for synset_id in wordnet_2.synsets():
        relations = wordnet_2.outbound_relations(synset_id)
        for relation in relations:
            try:
                if not wordnet_1.relation_exists(synset_id, relation[0], relation[1]):
                    diff_relations.add((synset_id, relation[1], relation[0]))
            except WordNetError:
                diff_relations.add((synset_id, relation[1], relation[0]))

    return diff_synsets if len(diff_synsets) > 0 else None, diff_relations if len(diff_relations) > 0 else None

